from . import AzureMLComputeCluster

__all__ = ["AzureMLComputeCluster"]

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
